#This mlp has following implementations
# - Weight balanced class sizes (error updated accordingly)
# - SMOTE (oversample minority class with noising)
# - oversample minority with replication, synthetic expansion
# - expand training module to make it more customizable, using lasagne
# - Add support for backtracking which features are the most important for each class
# - Add option for getting top-n classes predicted, not just the maximum prediction class

from __future__ import print_function
import os; os.environ["CUDA_VISIBLE_DEVICES"]="-1"
__docformat__ = 'restructedtext en'
import sys, timeit, gzip, argparse, json, shutil, csv, operator
from collections import Counter
import datetime

from itertools import repeat
import numpy as np; import math
from os.path import basename
import pandas as pd
import pickle, cPickle
from sklearn.utils import shuffle
import theano
import theano.tensor as T  
import glob

## Get dependent modules
from io_modules import Unbuffered, load_data, load_txt_input
from argtypes import restricted_float, proper_file, proper_dir, Dictionary, check_balance_type, proper_genecol, proper_delim
from normalization_functions import *
from sampling_algorithms import *
from calculate_stats import *
from plot_confusion import *
from split_input_into_testtrainvalid import *

#Lasagne functions
import lasagne
from lasagne.regularization import regularize_layer_params_weighted, l2, l1
from lasagne.regularization import regularize_layer_params
from calculate_nn_backprop import *
from lasagna_mlp_definitions import *
from sklearn.metrics import accuracy_score

# Has to be imported last, or segfault
import matplotlib
import matplotlib.pyplot as pyplot
matplotlib.use('Agg')

#genedict_file = os.path.abspath(sys.argv[0] + "/../../reference/gene_mapping_and_hugo_ids.txt")
def readtestinput(datfile, header_status, firstcol_ensg=True, transpose_dat=False, sep='\t'):
        "This function takes in a matrix of format samples (columns) x genes (rows) and generates an array of labels (Y) and an array of corresponding input vectors (X). The first column is the hugo ID, unless it is the ensemble ID followed by the hugo ID, in which case firstcol_ensg = True; the 3rd col onwards are the samples"
	print("Input has header? {0}, first two cols are gene names? {1}, setup is sample x gene? {2}, sep is: {3}".format(header_status, firstcol_ensg, transpose_dat, sep))
	if (datfile.endswith('.pkl')):
		with open(datfile, 'r') as infile:
			dataset=cPickle.load(infile)
			X=dataset[0]
			print(dataset[1])
			labels=dataset[1]
			Y=np.asarray(labels)
			samples=dataset[2]
	else:
		header_status = 0 if header_status == True else 'None'
		indat = pd.read_csv(datfile, sep=sep, header = header_status)
		if transpose_dat is True:
			indat2 = np.transpose(indat)
		else:
			indat2 = indat
		samples = list(indat2.columns)
		if firstcol_ensg is True:
			features_ensg = list(indat2.index); 
			features_hugo = list(indat2.iloc[:,0].values)
			sampleix = 1
		else:
			features_ensg = None
			features_hugo = list(indat2.index)
			sampleix = 0
		indat2 = indat2.replace("NA", float('nan'))
		X = indat2.iloc[:,sampleix:].values
		# Final X must be samples(rows) x genes(cols)
		if X.shape[0] == len(features_hugo):
			X = X.transpose()
	return [X, samples, features_ensg, features_hugo]


#Functions used in preprocessing (expanding dataset, or subsetting features based on an input list)
def make_blind(x, bfrac=0.0):
	blindarray = np.random.choice([0, 1], size=x.shape, p=[1.*bfrac, 1.0-bfrac])
	return(x*blindarray)

def adjust_trainset(x, y, adjust_type,verbose=False):
	print("Generating multiple iters using {0}...".format(adjust_type))
	duped_classx = []; duped_classy = []
	if adjust_type in ('smote', 'dup'):
		biggest_class_size = max(Counter(y).values())
		for ylab in set(y):
			class_indices = np.where(y == ylab)[0]
			subset_classx = np.take(x, class_indices,axis=0)
			classweight = float(biggest_class_size)/float(Counter(y)[ylab])
			#Extra samples required on top of duplicating by an integer factor
			spillover_samples = float(biggest_class_size) - (len(subset_classx)*np.floor(classweight))
			#expand the set
			duped_set = expand_this_set(minority_data_x = subset_classx, type=adjust_type, class_weight=classweight, additional_samples=spillover_samples)
			duped = (len(subset_classx)*np.floor(classweight))
			if verbose:
				print("CLASS {0}".format(ylab))
				print("....extending class {0} size {1} samples by factor of {2}, {3} samples".format(ylab, len(subset_classx), classweight, len(duped_set) ))
			duped_classx.extend(duped_set)
			duped_classy.extend(repeat(ylab, len(duped_set)))
	elif adjust_type == "imblearn":
		duped_classx, duped_classy = smote_this(x,y)
	elif adjust_type == "adasyn":
		duped_classx, duped_classy = adasyn_this(x,y)

	return(duped_classx, duped_classy)

def subset_features_x(x, fullfeatures_named, subsetfeatures_named):
	#Used to select a set of input features in a specific order
	#Format of input is samples x features
	subset_indices = []
	for select_this in subsetfeatures_named:
		subset_indices.append(fullfeatures_named.index(select_this))
	#returned object will have the features ordered in the manner of subsetfeatures_named
	return (np.apply_along_axis(lambda x: x[subset_indices], 1, x))

def padzero_features_x(x, fullfeatures_named, subsetfeatures_named, fillzero=True, default_value=0.0):
	reorder_subset = {subsetfeatures_named.index(i):fullfeatures_named.index(i) for i in subsetfeatures_named if i in fullfeatures_named}
        subset_keys_orderedix = [i[0] for i in sorted(reorder_subset.items(), key=operator.itemgetter(1))]
	subset_values = np.apply_along_axis(lambda x: x[subset_keys_orderedix], 1, x)
	set0_genes = []
        if fillzero:
                for i in range(0, len(fullfeatures_named), 1):
                        if fullfeatures_named[i] not in subsetfeatures_named:
                                subset_values=np.insert(subset_values, i, default_value, axis=1)
				set0_genes.append(fullfeatures_named[i])
        return(subset_values, set0_genes)

class basic_mlp_features(object):
	def __init__(self, learning_rate, L1_reg, L2_reg, n_epochs, dataset_file, batch_size, n_hidden, outdir, misclassflag, hiddenlayers=1, drop_input=0.0, drop_hidden=0.0,writeout=True,shuffle_train=False, normalize=None,normtype=None, keepsmall=False, test_dat = None, features_train=[]):
		self.inputfile = dataset_file
		self.outdir = outdir
		self.learning_rate=learning_rate; self.L1_reg = L1_reg; self.L2_reg = L2_reg
		self.n_epochs=n_epochs
		self.batch_size=batch_size
		self.n_hidden=n_hidden
		self.misclassflag=misclassflag
		self.writeout=writeout
		self.shuffle_train=shuffle_train
		self.rng = None
		#Params for data normalization
		self.normtype=normtype
		self.normalize=normalize
		#Param for data splitting
		self.keepsmall=keepsmall
		#Additional lasagna custom model inputs
		self.hiddenlayers=hiddenlayers
		self.drop_input=drop_input
		self.drop_hidden=drop_hidden
		self.separate_test=False
		if test_dat is not None:
			if(len(test_dat[0]) > 0):
				print("Testing input, assigning vars self.test_x after feat. selection and normalization.")
				self.X_test = np.asarray(test_dat[0])
				self.y_test = np.asarray(test_dat[1])
				self.separate_test=True
			
		#if features_train is not None:
		#if len(features_train) > 0:
		self.features_subset = features_train
		
	def load_mlp_data(self, fullset=False, verbose=True, shuffle_cv=False, save_idxs=False):
		if(self.inputfile.endswith(".txt")):
			self.load_txt_data(fullset=fullset, verbose=verbose, save_full_stats=save_idxs)
		else:
			self.load_pkl_data()
			print("Full set : {0}, saving idxs : {1}".format(fullset, save_idxs))
			if shuffle_cv:
				self.cv_shuffle_pkld_input(fullset=fullset,verbose=verbose, save_full_stats=save_idxs)
			elif fullset:
				self.cv_shuffle_pkld_input(fullset=fullset, verbose=verbose, save_full_stats=save_idxs)
		mlpobj.preprocess_data(verbose=True)

	def cv_shuffle_pkld_input(self, fullset, verbose=False, save_full_stats=False):
		print("Since input was pickled, pooling it back together and re-splitting. \n NO ADDITIONAL NORMALIZATION WILL BE DONE ON THE INPUT!")
		#Need to edit mydatstuff to return a pooled set, resplit, from input :(
		#Not sure if this dict is numeric_label:true_label key:value pair, or vice versa
		self.X_train, self.y_train = self.datasets[0]
		self.X_val, self.y_val = self.datasets[1]
		self.X_test_ds, self.y_test_ds = self.datasets[2]	
		self.x = np.concatenate((self.X_train, self.X_val, self.X_test_ds))
		self.y = np.concatenate((self.y_train, self.y_val, self.y_test_ds))
			
		self.labelnum = [self.labelnames.get(str(i)) for i in self.y]
		print(len(self.labelnum), len(self.y))
		dataseparator = mydatstuff(infile=None,x=self.x, labs=self.labelnum, ligs=self.features, outdir=self.outdir + "/dataset_train", trimflag=True, keepheader=True, drugbinderflag=False,keepsmall=self.keepsmall, fulltrainflag=fullset, classligands=None, testsize=0.2, outprefix="traincv", normtype=self.normtype, normby=self.normalize, makevalid=True,verbose=verbose)
		dataseparator.splitvalidtest()
		self.datasets, self.labelnames, self.labweighted, self.features = dataseparator.return_dataset()
		if save_full_stats:
			dataseparator.writeout() #Save idx's, not the actual dataset
			with open(os.path.join(self.outdir, 'featurelist.txt'), 'w') as f:
                                [ f.write('{0}\n'.format(x)) for x in self.features]					
		#Note that datasets[2], which is test, will be merged with validation if a test set has already been passed
		#This happens at training
			
	def load_pkl_data(self):
		self.datasets = load_data(self.inputfile, make_shared=False)
		indir = os.path.dirname(os.path.abspath(self.inputfile))
		self.labelweights = glob.glob(indir+"/dict_labweights_*")[0]
		self.labelset = glob.glob(indir+"/dict_labels_*")[0]
		self.featurelist = glob.glob(indir+"/featurelist.txt")
		self.labweighted = Dictionary(labelset=self.labelweights)
		self.labelnames = Dictionary(labelset=self.labelset)
		if self.featurelist:
			#print("...read in feature list")
			f = open(self.featurelist[0])
			with open(self.featurelist[0]) as f:
				self.features = [line.rstrip() for line in f]
		
		print("~~~~~LOADED PKL DATA FILE~~~~~~")	
	
	def load_txt_data(self,fullset=False,verbose=True, save_full_stats=False):
		dataseparator = mydatstuff(infile=self.inputfile, outdir=outdir, trimflag=True, keepheader=True, drugbinderflag=False,keepsmall=self.keepsmall, fulltrainflag=fullset, classligands=None, testsize=0.1, outprefix="testcv", normtype=self.normtype, normby=self.normalize, makevalid=True,verbose=verbose)
		dataseparator.splitvalidtest()
		dataseparator.normalize()
		self.datasets, self.labelnames, self.labweighted, self.features = dataseparator.return_dataset()
		if save_full_stats:
			dataseparator.writeout() #Save idx's, not the actual dataset
			
	def load_model(self, best_params_npz, label_dict_file, n_in, n_class):
		self.n_class = n_class; self.is_image=False
		print("Loading model from ... {0}".format(best_params_npz))
		with np.load(best_params_npz) as f:
			best_params = [f['arr_%d' % i] for i in range(len(f.files))]
		#set up network
		self.n_in = best_params[0].shape[0]; self.n_class = best_params[-1].shape[0]
		if self.n_in != n_in:
			print("Expect {0} and input {1}".format(self.n_in, n_in))
			print("WARNING: Passed input is not the same number of feats as trained model")
		self.target_var = T.ivector('targets') ; self.input_var = T.matrix('inputs') #Copied here from train during test run
			
		self.network = build_custom_mlp(input_var=self.input_var,n_out=self.n_class, num_features=self.n_in, depth=self.hiddenlayers, width=self.n_hidden, drop_input=self.drop_input, drop_hidden=self.drop_hidden, is_image=self.is_image)
		self.bestweights = [lasagne.layers.get_all_param_values(self.network)]	
		#Assign trained params to this network
		lasagne.layers.set_all_param_values(self.network, best_params)	
			
		#Update label dict mapping from integer to name
		self.labelnames = Dictionary(labelset=label_dict_file)
		if '0' in self.labelnames:
                        self.labelnames = dict([int(a),b] for a,b in self.labelnames.iteritems())
                else:
                        self.labelnames = dict([int(b),a] for a,b in self.labelnames.iteritems())
				
	def test_data(self, x_test=None, y_test=None):
		if x_test is None:
			x_test = self.X_test; y_test = self.y_test
		else:
			print("Input test set passed, different from runtime training-test sets")
		self.y_all_predicted = self.test_fn(x_test) #All label predictions for x
		self.y_predicted = np.argmax(self.y_all_predicted, axis=1) #Top label predicted for x
		self.y_predicted_probability = np.max(self.y_all_predicted, axis=1) #Probability score for top label 
		self.y_topn_predicted = map(lambda x: x[::-1][0:10], np.argsort(self.y_all_predicted, axis=1, kind='quicksort')) #top 10 labels predicted for x
		col_idx = np.arange(self.y_all_predicted.shape[0])[:,None]
		self.y_topn_predicted_probs=self.y_all_predicted[col_idx,self.y_topn_predicted]
		
	def write_test_output(self, samples=None):
		labelled_test_predict = ["c"] * len(self.y_predicted); labelled_test_predict_topn = ["c"] * len(self.y_predicted)
		for i in range(self.y_predicted.shape[0]):
			labelled_test_predict[i] = str(self.labelnames[self.y_predicted[i]])
			labelled_test_predict_topn[i] = map(lambda x:self.labelnames[x], self.y_topn_predicted[i])
		if not list(self.y_test):
			#If self.y_test is unassigned, re-assign it
			#Set it to a list of NAs
			self.y_test = ["NA"] * len(self.y_predicted)
		else:
			if isinstance(self.y_test[0], np.int):
                        	labelled_test_real = ["c"] * len(self.y_test);
				for i in range(self.y_test.shape[0]):
					labelled_test_real[i] = str(self.labelnames[self.y_test[i]])
			else:
				labelled_test_real = self.y_test
		if not not samples:
			#If a list of samples is provided, append that to self.y_test!
			self.y_test_labelled = zip(labelled_test_real, samples)
			header_all = list(["predicted","confidence","actual","sample"])
		else:
			header_all = list(["predicted","confidence","actual"])
					
		f_out=self.outdir + "/prediction_labels_ordered.txt"
		np.savetxt(f_out, np.column_stack((labelled_test_predict, self.y_predicted_probability, self.y_test_labelled)), fmt='%s',delimiter="\t")
		f_out2=self.outdir + "/prediction_labels_topn.txt"
		np.savetxt(f_out2, np.column_stack((self.y_test_labelled,  np.array(labelled_test_predict_topn))),delimiter="\t", fmt='%s')
		
		f_out3 = self.outdir+"/prediction_values_topn.txt"
		np.savetxt(f_out3, np.column_stack((self.y_test_labelled,np.array(self.y_topn_predicted_probs))),fmt='%s',delimiter="\t")
	
		fout=self.outdir+"/prediction_values_all.txt"
		np.savetxt(fout, np.column_stack(np.asarray(header_all+self.labelnames.values())), fmt='%s',delimiter="\t")
		with open(fout,'a') as f_out4:
			np.savetxt(f_out4, np.column_stack((labelled_test_predict, self.y_predicted_probability, self.y_test_labelled, np.array(self.y_all_predicted))), fmt='%s', delimiter="\t")
			
		print("TEST OUTPUT AT: {0}".format(self.outdir))
	
	def setup_test_data(self, is_train=False):
		#To asses testing, we don't need any dropout in the network
		#So we set deterministic=True (disables dropout)
		self.test_prediction = lasagne.layers.get_output(self.network, deterministic=True)
		self.test_loss = lasagne.objectives.categorical_crossentropy(self.test_prediction, self.target_var)
		self.test_loss = self.test_loss.mean()
		#if len(self.y_test)>0:
		#	self.test_acc = T.mean(T.eq(T.argmax(self.test_prediction, axis=1), self.target_var), dtype=theano.config.floatX)
	#		self.test_acc_fn = theano.function([self.input_var, self.target_var], [self.test_acc], allow_input_downcast=True)
		
		if is_train:
			if len(self.y_test)>0:
				self.test_acc = T.mean(T.eq(T.argmax(self.test_prediction, axis=1), self.target_var), dtype=theano.config.floatX)
				self.test_acc_fn = theano.function([self.input_var, self.target_var], [self.test_acc], allow_input_downcast=True)
				self.val_fn = theano.function([self.input_var, self.target_var], [self.test_loss ,self.test_acc], allow_input_downcast=True, on_unused_input='warn')
			else:
				print("WARNING: TRAINING data but no test set available for evaluation")
		
		self.test_fn = theano.function([self.input_var], lasagne.layers.get_output(self.network, deterministic=True), allow_input_downcast=True)
		
	def preprocess_data(self, verbose):
		if '0' in self.labelnames:
			self.labelnames = dict([int(a),b] for a,b in self.labelnames.iteritems())
		else:
			self.labelnames = dict([int(b),a] for a,b in self.labelnames.iteritems())

		self.sorted_keylabs=[]
		for i in sorted(self.labelnames.keys()):
                                self.sorted_keylabs.append(self.labelnames[i])
				
		self.X_train, self.y_train = self.datasets[0]
		self.X_val, self.y_val = self.datasets[1]
		
		#If test input if already passed, merge read-in dataset test component to validation
                if self.separate_test:
			print("TEST IS passed separately. Merging loaded test set with validation set.")
			self.X_val = np.concatenate((self.X_val, self.datasets[2][0]))
			self.y_val = np.concatenate((self.y_val, self.datasets[2][1]))
		else:
			print("TEST IS used from input")
			self.X_test, self.y_test = self.datasets[2] #NOT SURE IF THIS GOES HERE....
			
		print(self.X_train.shape, self.X_val.shape, self.X_test.shape)	
		if len(self.features_subset) > 0:
			self.X_train = subset_features_x(self.X_train, self.features, self.features_subset)
			self.X_val = subset_features_x(self.X_val, self.features, self.features_subset)
			self.X_test = subset_features_x(self.X_test, self.features, self.features_subset)
			self.features = self.features_subset
			print("___SUBSET INPUT, new training data shape {0}".format(self.X_train.shape))
			with open(os.path.join(self.outdir, 'featurelist.txt'), 'w') as f:
                                [ f.write('{0}\n'.format(x)) for x in self.features]
			
		self.n_class = len(set(self.y_train)) #Training set represents all the classes
		self.n_in = len(self.X_train[0]) #First input sample's length is the number of features (assumes it remains the same for all samples!)

		if isinstance(self.y_test[0], np.int):	
                        print("...--Verified test is correctly labelled--...")
		else:
			print("...--Test labels are alphabetical, mapping to numeric labels--...")
			#Temporarily invert labelnames dict so key:value is labelname:labelnum
			invert_dict = dict([b,a] for a,b in self.labelnames.iteritems())
			numeric_test_real = [0] * len(self.y_test)
			for i in range(self.y_test.shape[0]):
				if self.y_test[i] == "COAD_TS":
					self.y_test[i] = "COADREAD_TS"
				numeric_test_real[i] = invert_dict[self.y_test[i]]
			self.y_test = np.asarray(numeric_test_real) 
			
		if len(set(self.y_train)) != len(set(self.y_test)):
			print("WARNING: Test set does not have all the classes from the training set")
			print(set(self.y_train), set(self.y_test))
		if(verbose):
			print("Input is {0} \n Has {1} classes \n Learning params: \n {2} learning rate \n {3} l1 regression \n {4} l2 regression \n {5} hidden nodes \n {6} features per sample".format(self.inputfile, self.n_class, self.learning_rate, self.L1_reg, self.L2_reg, self.n_hidden, self.n_in))
		
        	if self.misclassflag == "none":
        	        print("No synthetic additions/modifications to the data")
        	elif self.misclassflag != "weight":
			self.X_train, self.y_train = adjust_trainset(self.X_train, self.y_train, self.misclassflag, verbose=verbose)

		if(self.shuffle_train):
                	print("...Shuffling training set")
			self.X_train, self.y_train  = shuffle(self.X_train, self.y_train, random_state=0)
		self.X_train = np.asarray(self.X_train)
		self.y_train = np.asarray(self.y_train)
		#print("Classes being trained:")
		#print(Counter(self.y_train))
		
	def get_classbased_features(self, writeout=False, outsuffix=""):
                if writeout:
                        #Create an output data directory
                        self.featoutdir=self.outdir + "/classbased_features" + outsuffix
                        try:
                                os.stat(self.featoutdir)
                        except:
                                os.mkdir(self.featoutdir)
                        print("...Writing class based features to output directory {0}".format(self.featoutdir))
                for i in range(self.y_predicted.shape[0]):
                        sample=self.y_test_labelled[i]
                        pred_class = self.y_predicted[i]                        
			correslist, featweights =get_weighted_features_per_class(pred_class, self.bestweights[0], feature_name_mapper=self.features)
                        #print("Class {0} influenced by features {1} ".format(self.labelnames[i],Counter(correslist)))
                        ##print("CLASS WEIGHINGS {0}".format(correslist))
                        if writeout:
                                with open(self.featoutdir + "/" + sample[1] + "_" + sample[0] + ".txt", 'w') as f:
                                        for labname, counts in Counter(correslist).most_common():
                                                f.write("{0}\t{1}\t{2}\n".format(sample,labname, counts))
                                with open(self.featoutdir + "/" + sample[1] + "_" + sample[0] + "weights.txt", 'w') as f:
                                        for i in range(len(featweights)):
                                                f.write("{0}\t{1}\t{2}\t{3}\n".format(sample[0],sample[1], featweights[i][0], featweights[i][1]))
		
	def get_performance_stats(self, y_test=None, y_pred=None, outdir=None, outsuffix=None, during_train=True):
		if y_test is None:
			y_test = self.y_test; 
		if y_pred is None:
			y_pred = self.y_predicted
		
		if not during_train:
			self.best_validation_loss=0; self.test_score=0
		
		if isinstance(y_test[0], np.int):
			labelled_test_real = ["c"] * len(y_test)
			for i in range(y_test.shape[0]):
				labelled_test_real[i] = str(self.labelnames[y_test[i]])
		else:
			labelled_test_real = y_test
			
		if isinstance(y_pred[0], np.int):
			labelled_test_predict = ["c"] * len(y_pred); 
			for i in range(y_pred.shape[0]):
				labelled_test_predict[i] = str(self.labelnames[y_pred[i]])
		else:
			labelled_test_predict = y_pred
		
		self.precision_test = calculate_precision(true_labs = labelled_test_real, pred_labs = labelled_test_predict)
		self.recall_test = calculate_recall(true_labs = labelled_test_real, pred_labs = labelled_test_predict)
		self.fscore_test = calculate_fbeta_score(true_labs = labelled_test_real, precision=self.precision_test, pred_labs = labelled_test_predict, recall=self.recall_test)
		self.mean_fscore = 1.00*sum(self.fscore_test.values()) / float(len(self.fscore_test.values()))
		if outdir is not None:
			try:
				self.sorted_keylabs
			except:
				self.sorted_keylabs = set(list(labelled_test_real) + list(labelled_test_predict))
			
			plot_confusion(pc=1, real_y=labelled_test_real,pred_y=labelled_test_predict, sorted_targetnames= self.sorted_keylabs,title="intermediate_test_output", outdir=outdir, outname="confusion"+outsuffix)
				
			pyplot.close()
		#Return accuracy on validation set, accuracy on test set, and fscore average on test set
		return(100.-(self.best_validation_loss * 100.), self.test_score * 100.,self.mean_fscore)
		#return(self.nn_best_val_acc, self.test_score * 100., self.mean_fscore)
			
def get_preds(myargs, modelname, modeldir):
	print("...TESTING input set")
	outdir=os.path.join(os.path.abspath(myargs.outdir[0]) + "/" + modelname + "_TESTPRED_" + myargs.pref[0] + "_" + datetime.datetime.now().strftime('%Y-%m-%d'));
	if not os.path.exists(outdir):
		os.makedirs(outdir)
	
	logfile = open(outdir + '/logfile.txt','w')
	sys.stdout = Unbuffered(sys.stdout, logfile=logfile)
                
	#Load the initialization parameters from training
	in_args = glob.glob(modeldir + "/argparse_args.txt")[0]
	in_model = glob.glob(modeldir + "/lasagne_bestparams.npz")[0]
	in_labdict = glob.glob(modeldir + "/dataset_train/"+ "*/dict_labels*"+".txt")[0]
	in_trainidx = glob.glob(modeldir + "/dataset_train/" + "*/trainingidxs.txt")[0]
	in_trainfeats = glob.glob(modeldir + "/featurelist.txt")[0]
	in_genecode = myargs.genecol[0];
	firstcol_ensg = myargs.multicolgenes
	in_delim = myargs.delim[0]
	in_transpose = myargs.transpose
	myargs_test = myargs #Update current myargs
	myargs = json.load(open(in_args)) #Retrieve old myargs to re-initialize lasagne model
	
	with open(in_trainfeats) as f:
		features_train = [line.rstrip() for line in f]
	
	#Step 1 : Prep input dataset
	print("UPDATE: Reading input from {0}".format(myargs_test.infile[0]))
	x_test, samples_test, features_ensg, features_test = readtestinput(datfile=myargs_test.infile[0], header_status=True, firstcol_ensg=firstcol_ensg, sep=in_delim, transpose_dat = in_transpose)
	print("UPDATE: Successfully read data, input shape is {0}, with {1} samples and {2} genes".format(x_test.shape, len(samples_test), len(features_test)))
	## If input FEATURES are not HUGO, map to HUGO from whichever format 
	if in_genecode not in ['HUGO', 'HUGO_ENSG']:
		print("UPDATE: Switching feature mapping to HUGO id")
		genedict = pd.read_csv(genedict_file, sep="\t")
		genedict = genedict.loc[genedict['HUGO'].isin(features_train)]
		genedict = genedict.set_index(in_genecode) # GENCODE, GAF, HUGO
		features_test = list(genedict.ix[features_test]["HUGO"])
	if in_genecode == "HUGO_ENSG":
		features_test = [m.split("_ENSG", 1)[0] for m in features_test]
	
	##First subselect features, ordered in order of training set
	print("UPDATE: We will be updating the gene set order, input features {0}, \n \t\t\t with training features {1}".format(features_test[0:10], features_train[0:10]))
        x_test, null_in_test = padzero_features_x(x_test, features_train, features_test)
        print("UPDATE: \n\t Genes not found in input: {0}".format(null_in_test))
	### Based on sea otter, so not using: x_test = subset_features_x(x_test, features_test, features_train)
       	print("UPDATE: Testing samples, samples x genes setup {0}".format(x_test.shape)) 
	if(len(samples_test) != x_test.shape[0]):
		samples_test  = ["unknown"] * x_test.shape[0]
	##Then normalize features according to training set
	myargs = argparse.Namespace(**myargs)
	if(myargs.normalize[0] == 1):
		print("UPDATE: Normalization is per sample, so not considering training dat")
		print("UPDATE: Normalization: {0}".format(myargs.normtype[0]))
		
		x_test = apply_norm_func(norm_func=myargs.normtype[0], bysamples=myargs.normalize[0], xdat=(x_test))
	else:
		print("Training based normalization is not valid")
		with open(in_trainidx) as f:
			train_idx = f.read().splitlines()
		#Do more here, idk what
		## subset train input by indices, and pass it along with test dat (x) to normalization function
	
	#Lastly, check if the flag for random noising/blinding is on:
	if(myargs_test.makeblind[0] > 0):
		print("UPDATE: Randomly setting {0} fraction of features to 0".format(myargs_test.makeblind[0]))
		x_test = make_blind(x_test, myargs_test.makeblind[0])
		
	#Step 2 : Initialize mlp obj
	mlpobj = basic_mlp_features(learning_rate=myargs.learn[0], L1_reg=myargs.l1reg[0], L2_reg=myargs.l2reg[0], n_epochs=myargs.epochs[0], n_hidden=myargs.hidden[0], dataset_file=myargs.input[0], batch_size=myargs.batches[0], outdir=outdir,misclassflag=myargs.classbalance[0],writeout=False,shuffle_train=myargs.shuffle, normtype=myargs.normtype[0],normalize=myargs.normalize[0], keepsmall=myargs.smallset, hiddenlayers=myargs.hiddenlayers[0], drop_input=myargs.dropinput[0], drop_hidden=myargs.drophidden[0], test_dat = [x_test, samples_test])
	mlpobj.load_model(best_params_npz=in_model, label_dict_file=in_labdict, n_in=x_test.shape[1],n_class=sum(1 for line in open(in_labdict)))
	
	#Step 3 : test stuffffff
	#print([x_test,y_test])
	print(all(isinstance(n, (int,long,float)) for n in x_test))
	mlpobj.setup_test_data(is_train=False)
	mlpobj.features = features_train
	mlpobj.test_data()
	#mlpobj.get_performance_stats(outdir=outdir, outsuffix="_testset", during_train=False)	
	mlpobj.write_test_output(samples=samples_test)
	mlpobj.get_classbased_features(writeout=True)
			

if __name__ == '__main__':
	myargs = getargs()
	v1_models_names = [item for item in dir(models_nold) if not item.startswith("__")]
	for model_name in v1_models_names:
		myargs = getargs()
		model_dir = getattr(models_nold, model_name)
		get_preds(myargs, model_name, model_dir)





import lasagne
import theano
import theano.tensor as T
import numpy as np
from lasagne.utils import floatX
from lasagne.regularization import regularize_layer_params_weighted, l2, l1
from lasagne.regularization import regularize_layer_params
from lasagne.objectives import squared_error as sqerror
from lasagne.layers import get_output as getout
import lasagne.layers 

#Define a custom autoencoder (called during training)
		
def cost_network(network, target_var, l1_reg, l2_reg, learn, train_layers=[], output_layer=[]):
	#for key in dont_train:
	#	network[key].params[network[key].W].remove("trainable")
	#	network[key].params[network[key].b].remove("trainable")
	#Basic loss is negative loss likelihood
	network_out = network[output_layer]
	prediction = lasagne.layers.get_output(network_out)
	loss = T.mean(T.nnet.categorical_crossentropy(prediction,target_var))
	#Shared costs
	l1_penalty = regularize_layer_params(network_out, l1) * l1_reg
	l2_penalty = regularize_layer_params(network_out, l2) * l2_reg
	cost = loss + l2_penalty + l1_penalty
	
	#params = lasagne.layers.get_all_params(network_out, trainable=True)
	#print(params)
	params=[]
	for p in train_layers:
		params.append(network[p].get_params(trainable=True))
	
	params = [item for sublist in params for item in sublist]
	
	print([i.eval().shape for i in params])
	print(params)	
	print(train_layers)
	print("----")
	updates = lasagne.updates.sgd(cost, params, learning_rate=learn)
	return([cost, updates, loss])
	
def build_deep_mlp(input_var=T.matrix('inputs'), target_var1=T.ivector('targets1'), target_var2=T.ivector('targets2'),target_var3=T.ivector('targets3'),target_var4=T.ivector('targets4'), num_features=17688, num_h1=12000, num_h2=10000, num_h3=8000, num_h4=4000, num_o1 = 2, num_o2 = 16, num_o3=40, num_o4=80, learn_rate=0.001, dropout_levels=[0,0,1,1,1],di=0.0, dh=0.5, is_image=False, l1_reg=0.0, l2_reg=0.0):
	#Assume last output layer is a multiclass output structure
	net = {}
	if is_image:
		net['l_in'] = lasagne.layers.InputLayer(shape=(None, 1, num_features, 1), input_var=input_var)
	else:
		net['l_in'] = lasagne.layers.InputLayer(shape=(None, num_features), input_var=input_var)
	
	nonlin = lasagne.nonlinearities.tanh #rectify
	nonlin_decode = lasagne.nonlinearities.softmax  #Multiclass output softmax
	init_weights = lasagne.init.GlorotUniform()
	bias = lasagne.init.Constant(0.)
	network_outputs = []; cost_outputs = []; updates_outputs = []
	train = []; predict=[]; val_fn=[]; test_acc_fn = []
	
	net['h1'] = lasagne.layers.DenseLayer(net['l_in'], num_units=num_h1, W=init_weights, nonlinearity=nonlin, b=bias)
	if(num_o1 != 0):
		net['o1'] = lasagne.layers.DenseLayer(net['h1'], num_units=num_o1, W=init_weights, nonlinearity=nonlin_decode, b=bias)
		cost1, updates1, base_loss = cost_network(net, target_var1, l1_reg, l2_reg, learn_rate, train_layers=["h1","o1"],output_layer="o1")
		network_outputs.append(net['o1']); cost_outputs.append(cost1); updates_outputs.append(updates1)
		train.append(theano.function([net['l_in'].input_var, target_var1], cost1, updates=updates1, allow_input_downcast=True))
		predict.append(theano.function([net['l_in'].input_var], lasagne.layers.get_output(net['o1'], deterministic=True), allow_input_downcast=True))
		predict_fn = theano.function([net['l_in'].input_var], lasagne.layers.get_output(net['o1'], deterministic=True), allow_input_downcast=True)
		test_acc = T.mean(T.eq(T.argmax(lasagne.layers.get_output(net['o1'], deterministic=True), axis=1), target_var1), dtype=theano.config.floatX)
		test_acc_fn.append(theano.function([net['l_in'].input_var, target_var1], [test_acc], allow_input_downcast=True))
		val_fn.append(theano.function([net['l_in'].input_var, target_var1], [base_loss, test_acc], allow_input_downcast=True, on_unused_input='warn'))
			
	####cost_outputs.append(sqerror(getout(o1), target_var_o1).mean() + regularize_layer_params(decoder_l_out, l2) * l2_reg)
	net['h2'] = lasagne.layers.DenseLayer(net['h1'], num_units=num_h2, W=init_weights, nonlinearity=nonlin, b=bias)
	
	net['o2']= lasagne.layers.DenseLayer(net['h2'], num_units=num_o2, W=init_weights, nonlinearity=nonlin_decode, b=bias)
	cost2, updates2, base_loss = cost_network(net, target_var2, l1_reg, l2_reg, learn_rate*1.0, train_layers=["h2","o2"], output_layer="o2")	
	network_outputs.append(net['o2']); cost_outputs.append(cost2); updates_outputs.append(updates2)
	train.append(theano.function([net['l_in'].input_var, target_var2], cost2, updates=updates2, allow_input_downcast=True))
	predict.append(theano.function([net['l_in'].input_var], lasagne.layers.get_output(net['o2'],deterministic=True), allow_input_downcast=True))
	test_acc = T.mean(T.eq(T.argmax(lasagne.layers.get_output(net['o2'], deterministic=True), axis=1), target_var2), dtype=theano.config.floatX)
	test_acc_fn.append(theano.function([net['l_in'].input_var, target_var2], [test_acc], allow_input_downcast=True))
	val_fn.append(theano.function([net['l_in'].input_var, target_var2], [base_loss, test_acc], allow_input_downcast=True, on_unused_input='warn'))
		
	net['h3'] = lasagne.layers.DenseLayer(net['h2'], num_units=num_h3, W=init_weights, nonlinearity=nonlin, b=bias)
	
	net['o3']= lasagne.layers.DenseLayer(net['h3'], num_units=num_o3, W=init_weights, nonlinearity=nonlin_decode, b=bias)
	cost3, updates3, base_loss = cost_network(net, target_var3, l1_reg, l2_reg, learn_rate*0.1, train_layers=["h2","h3","o3"], output_layer="o3")
	network_outputs.append(net['o3']); cost_outputs.append(cost3); updates_outputs.append(updates3)
	train.append(theano.function([net['l_in'].input_var, target_var3], cost3, updates=updates3, allow_input_downcast=True))
	predict.append(theano.function([net['l_in'].input_var], lasagne.layers.get_output(net['o3'],deterministic=True), allow_input_downcast=True))
	test_acc = T.mean(T.eq(T.argmax(lasagne.layers.get_output(net['o3'], deterministic=True), axis=1), target_var3), dtype=theano.config.floatX)
	test_acc_fn.append(theano.function([net['l_in'].input_var, target_var3], [test_acc], allow_input_downcast=True))
	val_fn.append(theano.function([net['l_in'].input_var, target_var3], [base_loss, test_acc], allow_input_downcast=True, on_unused_input='warn'))
	
	net['h4'] = lasagne.layers.DenseLayer(net['h3'], num_units=num_h4, W=init_weights, nonlinearity=nonlin, b=bias)
	
	net['o4'] = lasagne.layers.DenseLayer(net['h4'], num_units=num_o4, W=init_weights, nonlinearity=nonlin_decode, b=bias)
	cost4, updates4, base_loss = cost_network(net, target_var4, l1_reg, l2_reg, learn_rate*0.1, train_layers=["h1","h2","h3","h4","o4"], output_layer="o4")
	network_outputs.append(net['o4']); cost_outputs.append(cost4); updates_outputs.append(updates4)
	train.append(theano.function([net['l_in'].input_var, target_var4], cost4, updates=updates4, allow_input_downcast=True))
	predict.append(theano.function([net['l_in'].input_var], lasagne.layers.get_output(net['o4'],deterministic=True), allow_input_downcast=True))
	test_acc = T.mean(T.eq(T.argmax(lasagne.layers.get_output(net['o4'], deterministic=True), axis=1), target_var4), dtype=theano.config.floatX)
	test_acc_fn.append(theano.function([net['l_in'].input_var, target_var4], [test_acc], allow_input_downcast=True))
	val_fn.append(theano.function([net['l_in'].input_var, target_var4], [base_loss, test_acc], allow_input_downcast=True, on_unused_input='warn'))
	
	return network_outputs, train, predict, val_fn, test_acc_fn

#Define a custom mlp (called during training)
def build_custom_mlp(n_out, num_features, depth, width, drop_input, drop_hidden, input_var=None, is_image=False):
	if is_image:
		network = lasagne.layers.InputLayer(shape=(None, 1, num_features,1), input_var=input_var)
	else:
		network = lasagne.layers.InputLayer(shape=(None,num_features), input_var=input_var)
	if drop_input:
		network = lasagne.layers.dropout(network, p=drop_input)
	nonlin = lasagne.nonlinearities.tanh
	for _ in range(depth):
		network=lasagne.layers.DenseLayer(network, width, nonlinearity=nonlin)
		if drop_hidden:
			network = lasagne.layers.dropout(network, p=drop_hidden)
	#Final output layer is softmax, for multiclass output
	softmax = lasagne.nonlinearities.softmax
	network = lasagne.layers.DenseLayer(network, n_out, nonlinearity=softmax)
	return network


#Define iterator over minibatches of inputs (x) and corresponding labels (y)
def iterate_minibatches(inputs, targets, batchsize, shuffle=False):
	assert len(inputs) == len(targets)
	if shuffle:
		indices = np.arange(len(inputs))
		np.random.shuffle(indices)
	for start_idx in range(0, len(inputs) - batchsize + 1, batchsize):
		if shuffle:
			excerpt = indices[start_idx:start_idx + batchsize]
        	else:
			excerpt = slice(start_idx, start_idx + batchsize)
	yield inputs[excerpt], targets[excerpt]

#Define function to reshape 1 dimensional input to an image with 1 channel and width 1
def make_1d_image(x):
	return(x.reshape((x.shape[0],1,x.shape[1],1)))

class TransposedDenseLayer(lasagne.layers.DenseLayer):
	def __init__(self, incoming, num_units, W=lasagne.init.GlorotUniform(), b=lasagne.init.Constant(0.), nonlinearity=lasagne.nonlinearities.rectify,**kwargs):
		super(TransposedDenseLayer, self).__init__(incoming, num_units, W, b, nonlinearity, **kwargs)
	
	def get_output_shape_for(self, input_shape):
		return (input_shape[0], self.num_units)
	
	def get_output_for(self, input, **kwargs):
		if input.ndim > 2:
			input = input.flatten(2)
		activation = T.dot(input, self.W.T)
		if self.b is not None:
			activation = activation + self.b.dimshuffle('x', 0)
		return self.nonlinearity(activation)



import cancerscope
import os, sys
import tempfile

""" Need to bind raw_input to input in Python 2.* """
try:
	input = raw_input
except NameError:
	pass

def main():
	tempDir = "data/" #tempfile.mkdtemp()
	try:
		targetdir = input("Cancerscope requires pretrained models to be downloaded from the web. By default, these models will be downloaded to {0}.\nAlternatively, please enter a directory of your choice. A subdirectory named 'cancerscope_models' will be created there.\t".format(tempDir))
	except (SyntaxError):
		targetdir = tempDir
	
	""" raw_input() in python 2.* returns an empty string if user presses Return without input """
	if targetdir == "":
		targetdir = tempDir
	
	""" Now we can proceed with checking if the directory exists, and if not, raise error """
	if os.path.exists(targetdir):
		targetdir = targetdir + "/cancerscope_models/"
		if not os.path.exists(targetdir):
			os.mkdir(targetdir)
	else:
		print(targetdir is None)
		print("Please re-enter the target directory as {0} does not exist".format(targetdir))
		main(); return()
	
	print("Downloading data files to {0}".format(targetdir))
	cancerscope.get_models.getmodel(targetdir=targetdir)

if __name__ == "__main__":
	main()
	



"""
Setup file for models required by SCOPE
"""
from config import SCOPEMODELS_DATADIR, SCOPEMODELS_LIST, SCOPEMODELS_FILELIST_DIR
import cancerscope
import os, sys
import tempfile
import shutil, six
## Get detailed list of models and source files

modelOptions = cancerscope.getmodelsdict()

def findmodel(expected_pckgdir, model_label, expected_targetdir=None):
	expectedFilename = "model_" + model_label + ".txt"
	modelOptions_local = {}
	"""First check if the model details file exists in the main cancerscope directory (proof that model download attempt was completed atleast"""
	if os.path.exists(expected_pckgdir + "/" + expectedFilename):
		with open(expected_pckgdir + "/" + expectedFilename) as f:
			sys.stdout.write("\nReading model url and naming criterion for {0}...\n".format(model_label))
			for line in f:
				if line.strip()!= '':
					modelname_, url_, modeldir = line.strip().split('\t')
					modelOptions_local[modelname_] = modeldir + "/" + model_label.split("_")[-1]
	
	else:
		"""Otherwise, it would appear model wasnt even attempted to be downloaded"""
		if expected_targetdir is not None:
			sys.stdout.write("\nModel file {0} proceeding with download...\n".format(model_label))
			dnld_dir = downloadmodel(model_label = model_label, targetdir=expected_targetdir)
			modelOptions_local[model_label] = dnld_dir + "/" + model_label.split("_")[-1]
		else:
			sys.stdout.write("\nExpected model directory {0} does not exist...\n".format(expected_targetdir))
			return None
	
	if bool(modelOptions_local) is True:
		if os.path.exists(modelOptions_local[model_label] + "/lasagne_bestparams.npz") is True:
			sys.stdout.write("Model file found, returning to user...")
			return modelOptions_local
		else:
			if not os.path.exists(modelOptions_local[model_label]):
				"""The model directory does not exist in cancerscope/data, so need to re-download"""
				sys.stdout.write("\nModel download directory {0} does not exist...\n".format(modelOptions_local[model_label]))
				if expected_targetdir is not None:
					sys.stdout.write("\nDownloading model {0}".format(model_label))
					dnld_dir = downloadmodel(model_label = model_label, targetdir=expected_targetdir)
					modelOptions_local[model_label] = dnld_dir + "/" + model_label.split("_")[-1]
				else:
					sys.stdout.write("\nModel directory does not exist, but no target directory provided in case of missing model files. Return None...")
					return None
	else:
		sys.stdout.write("\nNo model details file found...")
		return None
	
def getmodel(model_label=None):
	"""Base function to retrieve models downloaded to package site directory"""
	model_dirs = {}
	if model_label is None:
		for m, _ in modelOptions.items():
			m_dirtemp = findmodel(expected_pckgdir = SCOPEMODELS_FILELIST_DIR, model_label = m, expected_targetdir = SCOPEMODELS_DATADIR)
			if m_dirtemp is not None:
				model_dirs[m] = m_dirtemp
			else:
				m_dirtemp = findmodel(expected_pckgdir = SCOPEMODELS_FILELIST_DIR, model_label = m)
	else:
		m_dirtemp = findmodel(expected_pckgdir = SCOPEMODELS_FILELIST_DIR, model_label = model_label)
		model_dirs[model_label] = m_dirtemp
	return model_dirs

def downloadmodel(model_label=None, targetdir=None):
	"""
	Query Zenodo and retrieve version-specific model parameter files and metadata
	"""
	global modelOptions
	if targetdir is not None:
		tempDir = targetdir
	else:
		tempDir = SCOPEMODELS_DATADIR ## By default, download models to the cancerscope/data directory, not tempfile.mkdtemp()
	
	if model_label is None:
		for m, _ in modelOptions.items():
			downloadmodel(m, tempDir)
	else:
		assert model_label in modelOptions.keys(), "%s is not a valid option in %s" % (model_label, modelOptions.keys())
		sys.stdout.write("\nDownloading model files for {0} \n\tData Downloaded at: {1}".format(model_label, tempDir))
		url, expectedFile, expectedmd5 = modelOptions[model_label]
		filesToDownload = [(url, expectedFile, expectedmd5)]
		expectedDir = expectedFile.replace(".tar.gz", "")
		
		if not os.path.isdir(tempDir):
			os.makedirs(tempDir)
		try:
			cancerscope.utils._downloadFiles(filesToDownload, tempDir)
		except Exception as e:
			print(e)
			exc_info = sys.exc_info(); shutil.rmtree(tempDir); six.reraise(*exc_info)
	
		mainDir = os.path.abspath(tempDir)
		with open(SCOPEMODELS_FILELIST_DIR + "/model_" + model_label + ".txt", "w") as f:
			f.write("\t".join([model_label, url, mainDir]))
		return(mainDir)




"""
Basic utility functions for cancerscope
"""

import os
import zipfile, tarfile
import hashlib
import requests
import logging
import traceback
import shutil

def file_as_bytes(file):
	with file:
		return file.read()

def _calcMD5(filename):
	return hashlib.md5(file_as_bytes(open(filename, 'rb'))).hexdigest()


# From: https://stackoverflow.com/questions/32763720/timeout-a-file-download-with-python-urllib
def _downloadFile(url, filename, timeout=180):
	request = requests.get(url, timeout=10, stream=True)
	with open(filename, 'w') as fh:
		## Walk through request response in chunks of 1024*1024 bytes
		for chunk in request.iter_content(1024*10240):
			fh.write(chunk)

def _downloadFiles(files, downloadDir):
	if not os.path.exists(downloadDir):
		os.mkdir(downloadDir)
	
	for url, shortName, expectedMD5 in files:
		downloadedPath = os.path.join(downloadDir, os.path.basename(url)) # Where will the compressed file be downloaded
		targetPathDir = os.path.join(downloadDir, shortName) ## What is the destination folder name for extracted files
		if os.path.isfile(downloadedPath):
			downloadedMD5 = _calcMD5(downloadedPath) 
			if not downloadedMD5 == expectedDM5:
				os.remove(downloadedPath)
		
		if not os.path.isfile(downloadedPath):
			try:
				_downloadFile(url, downloadedPath)
			except Exception as e:
				logging.error(traceback.format_exc())
				print(type(e)); raise
			downloadedMD5 = _calcMD5(downloadedPath)
			assert downloadedMD5 == expectedMD5, "MD5 checksum mismatch with downloaded file: %s" % shortName
			
			if downloadedPath.endswith('tar.gz'):
				tar = tarfile.open(downloadedPath, 'r:gz')
				tar.extractall(path=downloadDir)
				tar.close()
			os.remove(downloadedPath)




from __future__ import division
from sklearn.neighbors import NearestNeighbors
import numpy as np
import random
from itertools import repeat
#from imblearn import over_sampling
from collections import Counter

##Imblearn based functions
#def smote_this(all_x, all_y):
#	sm = over_sampling.SMOTE(kind='regular')
#	X_resampled, y_resampled = sm.fit_sample(all_x, all_y)
#	return X_resampled, y_resampled

def adjust_trainset(x, y, adjust_type, verbose=False):
	print("Generating synthetic samples using {0}...".format(adjust_type))
	duped_classx = []; duped_classy = []
	if adjust_type in ('smote', 'dup'):
		biggest_class_size = max(Counter(y).values())
		for ylab in set(y):
			class_indices = np.where(y == ylab)[0]
			subset_classx = np.take(x, class_indices, axis=0)
			classweight = float(biggest_class_size)/float(Counter(y)[ylab]) #Multiplicative term to expand smaller class by
			#Extra samples required, based on size of biggest class
			spillover_samples = float(biggest_class_size) - (len(subset_classx) * np.floor(classweight))
			#expand the set
			duped_set = expand_this_set(minority_data_x = subset_classx, type=adjust_type, class_weight=classweight, additional_samples=spillover_samples)
			duped = (len(subset_classx)*np.floor(classweight))
			if verbose:
				print("CLASS {0}".format(ylab))
				print("....extending class {0} size {1} samples by factor of {2}, {3} samples".format(ylab, len(subset_classx), classweight, len(duped_set) ))
			duped_classx.extend(duped_set)
			duped_classy.extend(repeat(ylab, len(duped_set)))
			
	elif adjust_type == "imblearn":
		duped_classx, duped_classy = probl_balance(x,y)
	elif adjust_type == "adasyn":
		duped_classx, duped_classy = adasyn_this(x,y)
	print("WARNING: Re-shuffle training set before learning, after set expansion")
	return(duped_classx, duped_classy)


##Shared functions
###probabilistic sampling
def probl_balance(x,y, dataset_size=None):
	classcounts = dict(Counter(y))
	largest_class = max(classcounts.values())
	classprobs = {k:1-(v/float(largest_class)) for k,v in classcounts.iteritems()}
	if dataset_size is None:
		dataset_size = x.shape[0]/len(set(y))
	xr = [] ; yr = []
	for i in range(0, x.shape[0]):
		rclass = random.randint(0,max(classcounts.keys())) #Randomly choose a class
		if random.random() < max(0.10,classprobs[rclass]):
			#If prob of selecting an element of this class is passed
			#Append an element from this  class to the list
			ridx = random.choice(np.where(y==rclass)[0])
			xr.append(x[ridx]); yr.append(y[ridx])
	
	return(np.asarray(xr), np.asarray(yr))
	#for i in range(0, x.shape[0]):
	#	if random.random() < max(0.10,classprobs[y[i]]):
	#		xr.append(x[i]); yr.append(y[i])

##ADASYN
#Implementation from pip install git+https://github.com/stavskal/ADASYN
#
def adasyn_this_test(all_x, all_y, tolerable_imbalance=0.9):
	from adasyn import ADASYN
	adsn = ADASYN(k=7, imb_threshold = tolerable_imbalance, ratio=0.75)
	new_X, new_y = adsn.fit_transform(all_x, all_y)
	return new_X, new_y

def adasyn_this(all_x, all_y, tolerable_imbalance=0.9):
	y_classes_count, sorted_y, cmax = getclasses(all_y)
	k=5
	return_np_array = all_x
	return_np_labs = all_y
	for class_small in sorted_y[1:]:
		csmall = y_classes_count[class_small]
		d_imb = csmall/cmax #Degree of imbalance
		if d_imb < tolerable_imbalance:
			#Set up the nearest neighbour model
			#NearestNeighbors will fit to the sample too, so we calculate k+1 neighbours and exclude current sample
			get_k_neighbors = NearestNeighbors(n_neighbors=k+1, algorithm='auto').fit(all_x)
			G_factor = (cmax - csmall) * 1
			csmall_indices = [i for i,x in enumerate(all_y) if x==class_small]
			small_list_r, small_list_neighbours = build_small_list(csmall_to_test = csmall_indices,csmall_indices=csmall_indices, get_k_neighbors=get_k_neighbors, all_x=all_x)
			#OPTIONAL: Set ri to 0 if only 1 out of 5 neighbours is a non-self example
			#print("Setting small ri (=<0.2) to 0")
		##	print(Counter(small_list_r))
			#small_list_r = [0 if x < 0.21 else x for x in small_list_r]
			#Normalize ri so that each ri corresponds to a distribution	
			#Calculate the number of new examples needed for each minority example xi,
			#By converting the ratios of (# neighbours in majority class / total neighbours) for each xi
			#Into a density distribution, and then converting it to required number of new examples
			#So you don't need new examples when the nearest neighbours are all the same class,
			# for all examples in a minority class
			if(sum(small_list_r) == 0):
				print("....All samples for class {0} define the class quite well, not generating new samples".format(class_small))
			else:
				normalize_ri = map(lambda x: (x/sum(small_list_r)), small_list_r);
				##Select G_factor based not on the extra samples needed to match input size, but to balance
				##Samples with all closest neighbours (normalize_ri == 0), with samples confounding to other classes
				##G_factor = min(G_factor, len([x for x in small_list_r if x ==0.0]))
				normalize_ri = map(lambda x: (x/sum(small_list_r)), small_list_r);
				gi = map(lambda x: np.ceil(G_factor * x), normalize_ri)
		##		print("...Generating {0} samples for class {1}, needed {2}".format(sum(gi),class_small, G_factor))
				for i in range(0, len(csmall_indices)):
					ci = csmall_indices[i]
					cxi = small_list_neighbours[i]
					#If there are no nearest neighbours in the same class for a sample
					#Then assign this sample as its own nearest neighbour
					if(gi[i] > 0):
						if(len(cxi) == 0): 
							cxi=ci ; #print(ci, cxi, gi[i])
						random_k_idx = np.random.choice(cxi, gi[i], replace=True)
						for kix in random_k_idx:
							gap = random.betavariate(1,1)
							si = all_x[ci] + (all_x[kix] - all_x[ci])*gap
							return_np_array = np.append(return_np_array, si.reshape(1,-1), axis=0)
							return_np_labs = np.append(return_np_labs, class_small.reshape(1,-1))
		##	print("...CLASS {0} extended to size {1}".format(class_small, len(return_np_labs[return_np_labs==class_small])))
			
	for class_small in sorted_y[1:]:
			csmall_indices = [i for i,x in enumerate(all_y) if x==class_small]
			get_k_neighbors = NearestNeighbors(n_neighbors=k+1, algorithm='auto').fit(all_x)
			small_list_r, small_list_neighbours = build_small_list(csmall_to_test = csmall_indices, csmall_indices=csmall_indices,get_k_neighbors=get_k_neighbors, all_x=all_x)
			print(Counter(small_list_r))
			get_k_neighbors = NearestNeighbors(n_neighbors=k+1, algorithm='auto').fit(return_np_array)
			new_csmall_indices = [i for i,x in enumerate(return_np_labs) if x==class_small] 
			small_list_r, small_list_neighbours = build_small_list(csmall_to_test = csmall_indices, get_k_neighbors=get_k_neighbors, all_x=all_x, csmall_indices = new_csmall_indices)
		##	print(Counter(small_list_r))
		##	print(".....END OF CLASS{0}......".format(class_small))
	return return_np_array, return_np_labs

def build_small_list(csmall_to_test, get_k_neighbors, all_x, csmall_indices):
	#This function takes a list of indices indicating the location of the small class
	#And a pre-fit NearestNeighbours function
	#And generates a list of ratios indicating, for each sample,(number of non-self neighbours)/(total neighbours)
	#Also returns the actual list of neighbours for each sample
	small_list_neighbours = []; small_list_r = []
	for ci in csmall_to_test:
		xi = all_x[ci].reshape(1,-1)
		indices = getknn(get_k_neighbors, xi)
		ci_neighbours = [x for x in indices if x in csmall_indices]
		small_list_r.append( float(len(indices) - len(ci_neighbours))/len(indices) )
		small_list_neighbours.append(ci_neighbours)
	return(small_list_r, small_list_neighbours)

def getknn(fit_knn, sample, exclude_self=True):
	#Assumes fit_knn was fit including the sample, so we exclude it from the fit's NearNeighbor results
	dist, ind = fit_knn.kneighbors(sample)
	if exclude_self:
		return(ind[0][1:])
	else:
		return(ind[0])

def getclasses(y):
	y_classes_count = Counter(y)
	sorted_y = sorted(y_classes_count, key=y_classes_count.get, reverse=True)
	cmax = y_classes_count[sorted_y[0]]
	return [y_classes_count, sorted_y, cmax]
	
##SMOTE 
def expand_this_set(minority_data_x, type,  additional_samples=0, class_weight = 0):
        #minority_data_x : nparray of samples with set of features representing each sample
        #It is assumed that the passed data is all from the same (minority) class
        #For k, We will pick either the 5 nearest neighbours, or the size of the input class 
	#Whichever is lesser
	#inflate_factor is 2^class weight, calculated as (biggest_class_size/this_class_size)
	return_np_array = minority_data_x #Get the first set of values for this class as the default samples
	additional_samples = int(additional_samples)
	inflate_factor = int(np.floor(class_weight) - 1)
	if type=="smote":
		#WARNING: Currently not excluding self from set of nearest neighbours when you have a small class,
		# so it can be chosen as the nearest neighbour
		class_size = len(minority_data_x)
		min_neighbours = int(min(5, class_size))
		get_k_neighbors = NearestNeighbors(n_neighbors=min_neighbours, algorithm='auto').fit(minority_data_x)
		#Plus select random minority samples to fill up the additional < (len(minority_set)) samples that are needed
		if inflate_factor > 0:
		##	print("....Synthetically extending dataset {0} time".format(inflate_factor))
			for x in minority_data_x:
				x = x.reshape(1,-1)
				indices = getknn(get_k_neighbors,x, exclude_self=(class_size > min_neighbours))
				#Uniform random sampling to pick a random sample from the k nearest neighbours
				random_k_idx = np.random.choice(indices, inflate_factor, replace=(inflate_factor > len(indices)))
				diff = x - minority_data_x[random_k_idx]
				gap = random.betavariate(1,1); n = x + diff*gap
				return_np_array = np.append(return_np_array, n, axis=0)
		if additional_samples >0:
		##	print("....artifically generating {0} additional samples".format(additional_samples))
			addlist = np.asarray([random.choice(minority_data_x) for _ in range(additional_samples)])
			for x in addlist:
				x = x.reshape(1,-1)
				indices = getknn(get_k_neighbors, x, exclude_self = True)
				random_k_idx = np.random.choice(indices, 1, replace=False)
				diff = x - minority_data_x[random_k_idx]
				gap = random.betavariate(1,1); n = x + diff*gap
				return_np_array = np.append(return_np_array, n, axis=0)	
	if type=="dup":
		return_np_array = np.repeat(minority_data_x, inflate_factor, axis=0)
		#Lastly, add the additional samples from the original by random sampling
		if(additional_samples > 0):
		##	print("....randomly sampling {0} additional samples on top of extending samples (synthetically or otherwise) {1} times".format(additional_samples, inflate_factor-1))
			addlist = np.asarray([random.choice(minority_data_x) for _ in range(additional_samples)])
			return_np_array = np.concatenate((return_np_array, addlist),axis=0)	
	
	return return_np_array




#GITHUB source https://gist.github.com/ebenolson/1682625dc9823e27d771
#Draw neural net; need graphviz and pydot (in that order)
"""
Functions to create network diagrams from a list of Layers.

Examples:

    Draw a minimal diagram to a pdf file:
        layers = lasagne.layers.get_all_layers(output_layer)
        draw_to_file(layers, 'network.pdf', output_shape=False)

    Draw a verbose diagram in an IPython notebook:
        from IPython.display import Image #needed to render in notebook

        layers = lasagne.layers.get_all_layers(output_layer)
        dot = get_pydot_graph(layers, verbose=True)
        return Image(dot.create_png())
"""

import pydot


def get_hex_color(layer_type):
    """
    Determines the hex color for a layer. Some classes are given
    default values, all others are calculated pseudorandomly
    from their name.
    :parameters:
        - layer_type : string
            Class name of the layer

    :returns:
        - color : string containing a hex color.

    :usage:
        >>> color = get_hex_color('MaxPool2DDNN')
        '#9D9DD2'
    """

    if 'Input' in layer_type:
        return '#A2CECE'
    if 'Conv' in layer_type:
        return '#7C9ABB'
    if 'Dense' in layer_type:
        return '#6CCF8D'
    if 'Pool' in layer_type:
        return '#9D9DD2'
    else:
        return '#{0:x}'.format(hash(layer_type) % 2**24)


def get_pydot_graph(layers, output_shape=True, verbose=False):
    """
    Creates a PyDot graph of the network defined by the given layers.
    :parameters:
        - layers : list
            List of the layers, as obtained from lasange.layers.get_all_layers
        - output_shape: (default `True`)
            If `True`, the output shape of each layer will be displayed.
        - verbose: (default `False`)
            If `True`, layer attributes like filter shape, stride, etc.
            will be displayed.
        - verbose:
    :returns:
        - pydot_graph : PyDot object containing the graph

    """
    pydot_graph = pydot.Dot('Network', graph_type='digraph')
    pydot_nodes = {}
    pydot_edges = []
    for i, layer in enumerate(layers):
        layer_type = '{0}'.format(layer.__class__.__name__)
        key = repr(layer)
        label = layer_type
        color = get_hex_color(layer_type)
        if verbose:
            for attr in ['num_filters', 'num_units', 'ds',
                         'filter_shape', 'stride', 'strides', 'p']:
                if hasattr(layer, attr):
                    label += '\n' + \
                        '{0}: {1}'.format(attr, getattr(layer, attr))
            if hasattr(layer, 'nonlinearity'):
                try:
                    nonlinearity = layer.nonlinearity.__name__
                except AttributeError:
                    nonlinearity = layer.nonlinearity.__class__.__name__
                label += '\n' + 'nonlinearity: {0}'.format(nonlinearity)

        if output_shape:
            label += '\n' + \
		'Output shape: {0}'.format(layer.output_shape)
  #              'Output shape: {0}'.format(layer.get_output_shape())
        pydot_nodes[key] = pydot.Node(key,
                                      label=label,
                                      shape='record',
                                      fillcolor=color,
                                      style='filled',
                                      )

        if hasattr(layer, 'input_layers'):
            for input_layer in layer.input_layers:
                pydot_edges.append([repr(input_layer), key])

        if hasattr(layer, 'input_layer'):
            pydot_edges.append([repr(layer.input_layer), key])

    for node in pydot_nodes.values():
        pydot_graph.add_node(node)
    for edge in pydot_edges:
        pydot_graph.add_edge(
            pydot.Edge(pydot_nodes[edge[0]], pydot_nodes[edge[1]]))
    return pydot_graph


def draw_to_file(layers, filename, **kwargs):
    """
    Draws a network diagram to a file
    :parameters:
        - layers : list
            List of the layers, as obtained from lasange.layers.get_all_layers
        - filename: string
            The filename to save output to.
        - **kwargs: see docstring of get_pydot_graph for other options
    """
    dot = get_pydot_graph(layers, **kwargs)

    ext = filename[filename.rfind('.') + 1:]
    with open(filename, 'w') as fid:
        fid.write(dot.create(format=ext))


def draw_to_notebook(layers, **kwargs):
    """
    Draws a network diagram in an IPython notebook
    :parameters:
        - layers : list
            List of the layers, as obtained from lasange.layers.get_all_layers
        - **kwargs: see docstring of get_pydot_graph for other options
    """
    from IPython.display import Image  # needed to render in notebook

    dot = get_pydot_graph(layers, **kwargs)
    return Image(dot.create_png())



import os
import argparse

#Define basic params used during execution
def roundup(x):
	return int(math.ceil(x / 1000.0)) * 1000

def restricted_float(x):
        x = float(x)
        if x < 0.0 or x > 1.0:
                raise argparse.ArgumentTypeError("%r not i nrange [0.0, 1.0]"%(x,))
        else:
                return x

def proper_delim(x):
	delims_list = ('\t', ' ', ',')
	if not x in delims_list:
		raise argparse.ArgumentTypeError("%r is not a valid Delimiter" % (x, ))
		raise argparse.ArgumentTypeError('Possible values {0}'.format(delims_list))
	else:
		return x.upper()

def proper_genecol(x):
	genecol_list = ('HUGO', 'GAF', 'GENCODE', 'ENSG', 'HUGO_ENSG')
	if not x.upper() in genecol_list:
		raise argparse.ArgumentTypeError("%r is not a valid Gene Name type " % (x, ))
		raise argparse.ArgumentTypeError('Possible values {0}'.format(genecol_list))
	else:
		return x.upper()

def proper_file(x):
	x=os.path.abspath(x)
        if not os.path.exists(x):
                raise argparse.ArgumentTypeError("%r is not a valid file"%(x,))
        else:
                return x

def proper_dir(x):
        if not os.path.exists(x):
                if not os.path.exists(os.path.join(x,"/../")):
                        raise argparse.ArgumentTypeError("%r is not a valid dir"%(x,))
                else:
                        os.makedirs(x)
                        return x
        else:
                return x

def Dictionary(labelset):
    with open(labelset) as f: d = dict((k.rstrip(), v.rstrip()) for k,v in (line.split('\t') for line in f))
    return (d)

def makeintifpossible(x):
    try:
        y = int(float(x))
    except:
        y = float('nan')

    return y

def int_or_float(x):
    x = float(x)
    if x < 0.0 or x > 1.0:
        x = int(x)
        if x > 1:
            return x
        raise argparse.ArgumentTypeError('$r is not a float in range [0,1], nor an int > 1' % (x,))
    else:
        return x

def roundup(x):
	return int(match.ceil(x / 1000.0)) * 1000

def check_norm_values(x):
    normlist = ('minmax', 'scale', 'normscale','none', 'rasterize','rastminmax')
    if x not in normlist:
        raise argparse.ArgumentTypeError('%r is not a valid normalization method. ' % (x,))
	raise argparse.ArgumentTypeError('Possible values {0}'.format(normlist))
    else:
        return x

def check_balance_type(x):
        if x not in ('none', 'weight', 'smote', 'dup','imblearn', 'adasyn'):
                raise argparse.ArgumentTypeError('%r is not a valid class expansion method. Possible values - none, weight, smote, dup , adasyn, imblearn' %(x,))
        else:
                return x




from __future__ import print_function

def readinput(datfile, header_status, firstcol_ensg=True, transpose_dat=False, sep='\t'):
        "This function takes in a matrix of format samples (columns) x genes (rows) and generates an array of labels (Y) and an array of corresponding input vectors (X). The first column is the hugo ID, unless it is the ensemble ID followed by the hugo ID, in which case firstcol_ensg = True; the 3rd col onwards are the samples"
	print("Input has header? {0}, first two cols are gene names? {1}, setup is sample x gene? {2}, sep is: {3}".format(header_status, firstcol_ensg, transpose_dat, sep))
	if (datfile.endswith('.pkl')):
		with open(datfile, 'r') as infile:
			dataset=cPickle.load(infile)
			X=dataset[0]
			print(dataset[1])
			labels=dataset[1]
			Y=np.asarray(labels)
			samples=dataset[2]
	else:
		header_status = 0 if header_status == True else 'None'
		indat = pd.read_csv(datfile, sep=sep, header = header_status)
		if transpose_dat is True:
			indat2 = np.transpose(indat)
		else:
			indat2 = indat
		samples = list(indat2.columns)
		if firstcol_ensg is True:
			features_ensg = list(indat2.index); 
			features_hugo = list(indat2.iloc[:,0].values)
			sampleix = 1
		else:
			features_ensg = None
			features_hugo = list(indat2.index)
			sampleix = 0
		indat2 = indat2.replace("NA", float('nan'))
		X = indat2.iloc[:,sampleix:].values
		# Final X must be samples(rows) x genes(cols)
		if X.shape[0] == len(features_hugo):
			X = X.transpose()
	return [X, samples, features_ensg, features_hugo]


	## If input FEATURES are not HUGO, map to HUGO from whichever format 
	if in_genecode not in ['HUGO', 'HUGO_ENSG']:
		print("UPDATE: Switching feature mapping to HUGO id")
		genedict = pd.read_csv(genedict_file, sep="\t")
		genedict = genedict.loc[genedict['HUGO'].isin(features_train)]
		genedict = genedict.set_index(in_genecode) # GENCODE, GAF, HUGO
		features_test = list(genedict.ix[features_test]["HUGO"])
	if in_genecode == "HUGO_ENSG":
		features_test = [m.split("_ENSG", 1)[0] for m in features_test]





"""
Setup script for SCOPE (Python package: 'cancerscope')
"""
from setuptools import setup, find_packages
import os, sys
import re

"""Identify basic variables needd for setup"""
pckg_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(pckg_dir, 'cancerscope/__init__.py'), 'r') as f:
	VERSION = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open(os.path.join(pckg_dir, 'requirements.txt'), 'r') as f:
	requirements = f.readlines()

with open(os.path.join(pckg_dir, 'README.md'), 'r') as f:
	longdescription = f.read()

""" Post-setup script to download models 
As per answer posted at https://stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools
""" 

""" Setup instructions """ 
setup(name='cancerscope', 
	version=VERSION, 
	setup_requires=['nose>=1.0'],
	description='An RNA-Seq based tool for Supervised Cancer Origin Prediction using Expression',
	long_description=longdescription, 
	long_description_content_type='text/markdown',
	author='Jasleen Grewal',
	author_email='grewalj23@gmail.com', 
	url='https://pypi.org/project/cancerscope/', 
	packages=find_packages(exclude=['tests*']), 
	license='MIT', 
	#dependency_links=['https://github.com/Lasagne/Lasagne/tarball/master#egg=lasagne-0.2.dev1', 'https://github.com/Theano/Theano/tarball/master#egg=theano-0.8'],
	python_requires='>=2.6, <=3.5',
	install_requires=requirements, 
	include_package_data=True, zip_safe=False, 
	test_suite='nose.collector', tests_require=['nose'], 
	classifiers = ['Programming Language :: Python :: 2.7', 'Topic :: Scientific/Engineering :: Artificial Intelligence', 'Development Status :: 5 - Production/Stable', 'Intended Audience :: Healthcare Industry', 'Topic :: Scientific/Engineering :: Medical Science Apps.', 'Topic :: Scientific/Engineering :: Bio-Informatics'], 
	package_data = {'cancerscope': ['resources/*.txt', '*.rst']} 
)




import numpy as np
from itertools import chain
from collections import Counter 

#Get predicted label
def get_weighted_features_per_class(class_numlabel, modelset, feature_name_mapper=None):
	#Class_numlabel ranges from 0 to max_num_classes-1; is an int
        #modelset is of format [W_1stlayerfrominput, bias_1stlayerfrominput, ..., W_layerbeforeoutput, bias_layerbeforeoutput]
        #feature_name_mapper is a list of length num_input_features with ordered with input names corresponding to input feature index. Izzz optional
        ypred = class_numlabel
        #Now, work backwards, assuming modelset = [W, b, W, b]
        class_bias=[]; class_weights_perlayer=[];
	is_nparray = isinstance(modelset[0], (np.ndarray,np.generic))	
        if is_nparray:
		for i in range(0,len(modelset)/2):
			class_bias.append(modelset[(i*2)+1])
			class_weights_perlayer.append(modelset[(i*2)])
	else:
		for i in range(0,len(modelset)/2):
			class_bias.append(np.array(modelset[(i*2)+1].eval()))
			class_weights_perlayer.append(np.array(modelset[(i*2)].eval()))
	
	#Forward multiplication guys!
	for i in range(1, len(modelset)/2):
		if i==1:
			carry_forward = np.dot((class_weights_perlayer[0] + class_bias[0]), (class_weights_perlayer[i] + class_bias[i]))
		else:
			carry_forward = np.dot(carry_forward, (class_weights_perlayer[i] + class_bias[i]))
	#Final matrix (carry_forward) is num_features x num_classes, so figure out class specific topn features
	top_features_queryclass = np.transpose(carry_forward)[ypred].argsort()[::-1]#[0:100]
	values_features_queryclass = np.transpose(carry_forward)[ypred]
	if feature_name_mapper is not None:
		feature_named = []
		featweights = []
		for feat_ix in top_features_queryclass:
			featweights.append([feature_name_mapper[feat_ix], values_features_queryclass[feat_ix]])
			
			feature_named.append(feature_name_mapper[feat_ix])
	else:
		feature_named = top_features_queryclass
		
	return (feature_named, featweights)
	
def get_weighted_features_per_class_eachstep(class_numlabel, modelset, feature_name_mapper=None):
	#Class_numlabel ranges from 0 to max_num_classes-1
	#modelset is of format [W_1stlayerfrominput, bias_1stlayerfrominput, ..., W_layerbeforeoutput, bias_layerbeforeoutput]
	#feature_name_mapper is a list of length num_input_features with ordered with input names corresponding to input feature index. Izzz optional
	ypred = class_numlabel

	#Now, work backwards, assuming modelset = [W, b, W, b]
	class_bias=[]; class_weights_perlayer=[]; 
	for i in range(0,len(modelset)/2):
		class_bias.append(np.array(modelset[(i*2)+1].eval()))
		class_weights_perlayer.append(np.array(modelset[(i*2)].eval()))
		#class_bias.append(modelset[(i*2)+1])
		#class_weights_perlayer.append(modelset[(i*2)])	
	#Now, presuming input is inputx, figure out which of the input features were weighted the highest
	topn=10
	top_weighted_iminus1thlayer=[]
	for i in range(0, len(class_bias))[::-1]:
		if i == len(class_bias) - 1 :
			#If we are at the instance of initial backprop, claculate top weighted nodes in the out-1 layer
			top_weighted_iminus1thlayer.append((np.transpose(class_weights_perlayer[i])[ypred]).argsort()[::-1][0:topn].tolist())
			top_weighted_currentminusone=list(chain.from_iterable(top_weighted_iminus1thlayer[-1:]))
		else:
			#print("Accessing {0} + 1th layer from initial input layer".format(i))
			#print("...its highest ranked weights are {0}".format(top_weighted_currentminusone))
			templist=[]
			for top_node in np.unique(top_weighted_currentminusone):
				templist.append((np.transpose(class_weights_perlayer[i])[top_node]).argsort()[::-1][0:topn].tolist())
			top_weighted_iminus1thlayer.append(templist)
			top_weighted_currentminusone=list(chain.from_iterable(top_weighted_iminus1thlayer[-1]))
		
	if feature_name_mapper is not None:
	#the last output from the loop was the actual most weighted features, so we use that
		feature_numeric=top_weighted_currentminusone
		feature_named = []
		for feat_ix in feature_numeric:
			feature_named.append(feature_name_mapper[feat_ix])
	else:
		feature_named = top_weighted_currentminusone
	
	return(top_weighted_iminus1thlayer, feature_named)

	



""" 
cancerscope provides cancer-type predictions for RNA-Seq samples  
"""
import os, sys
from tests import *

### Global opts
from config import REF_LABELCOUNTS,SCOPEMODELS_DATADIR, SCOPEMODELS_LIST, SCOPEMODELS_FILELIST_DIR, REF_DISEASECODES, getmodelsdict
from cancerscope.utils import *

### Data sources
from cancerscope.get_models import getmodel, findmodel, downloadmodel
from cancerscope.get_models import *

### Dependent data
from cancerscope.scope_plots import *
from cancerscope.scope_ensemble import scopemodel, scope

if sys.version_info < (2, 7):
	raise ImportError("Cancerscope module requires Python 2.7 or higher")

__version__ = '0.29'



